package se.umu.jobsubmission.impl.broker; 

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.apache.axis.message.MessageElement;
import org.apache.log4j.Logger;
import org.ggf.schemas.jsdl._2005._11.jsdl.JobDefinition_Type;
import org.globus.gsi.GlobusCredential;
import org.globus.mds.glue.BenchmarkType;
import org.globus.mds.glue.ClusterType;
import org.globus.mds.glue.GLUECERPType;
import org.globus.mds.glue.SubClusterType;
import org.globus.wsrf.encoding.ObjectDeserializer;

import se.umu.jobsubmission.Benchmark;
import se.umu.jobsubmission.BenchmarkList;
import se.umu.jobsubmission.JobPreferencesType;
import se.umu.jobsubmission.SchedulingHintsType;

import se.umu.jobsubmission.impl.dispatcher.Dispatcher; 

public class RunTimePredictor implements Predictor {
	
	private static Logger logger = Logger.getLogger(RunTimePredictor.class); 
	
	/**
	 * Extracts benchmarks advertised by cluster and adds them to a map. 
	 * Currently, only specint2000 and specfp2000 are supported. 
	 * @param resource
	 * @return
	 * @throws Exception
	 */
	public Map getClusterBenchmarks(GLUECERPType resource) throws Exception {
		
		Map results = new HashMap(); 
		
		try {
			
			ClusterType cluster = resource.getCluster(0); 
			BenchmarkType arcBenchmarks = cluster.getSubCluster(0).getHost(0).getBenchmark();  
			
			if(arcBenchmarks == null) {
				throw new Exception("No benchmarks found for " + 
						resource.getCluster(0).getSubCluster(0).getHost(0).getName()); 
			}
			
			MessageElement any[] = arcBenchmarks.get_any(); 
	    	BenchmarkList glueBenchmarkList = (BenchmarkList) 
					ObjectDeserializer.toObject(any[0], BenchmarkList.class); 
	    	
	    	Benchmark glueBenchmarks[] = glueBenchmarkList.getBenchmarks(); 
	    	for(int i = 0; i < glueBenchmarks.length; i++) {
	    		logger.debug("Benchmark: " + glueBenchmarks[i].getName() 
	    				+ " @ " + glueBenchmarks[i].getResult()); 
	    		results.put(glueBenchmarks[i].getName(), 
	    				new Float(glueBenchmarks[i].getResult())); 
	    	}
			
		} catch (Exception e) {
			logger.debug("Error determining cluster benchmarks"); 
			throw e; 
		}
		
		return results; 
			
	}
	
	/**
	 * Performs benchmark-based runtime prediction. 
	 * @param resource - GLUE info about resource, including benchmark results.
	 * @param benchmarks - Benchmarks specified by user. 
	 * @return
	 * @throws Exception
	 */
	public int benchmarkRunTime(GLUECERPType resource, 
			Benchmark[] benchmarks) throws Exception {		
				
		logger.debug("Determining run time from benchmarks"); 
		
		Map clusterBenchmarks; 
		
		// Get benchmarks, Exception is thrown if none are found. 
		clusterBenchmarks = getClusterBenchmarks(resource); 
		
		int noMissing = 0; 
		List clusterTimes = new Vector(); 
		String userBenchmarkName; 
		float userBenchmarkResult; 
		int userBenchmarkTime; 
		
		logger.debug("Cluster benchmarks are: " + clusterBenchmarks); 
		
		// Estimate the runtime separately for each user-specified benchmark.
		// Save runtime estimates in a list and note missing
		// benchmarks. 
		for(int i = 0;  i < benchmarks.length; i++) {
			userBenchmarkName = benchmarks[i].getName(); 
			userBenchmarkResult = benchmarks[i].getResult(); 
			userBenchmarkTime = benchmarks[i].getTime(); 
			
			logger.debug("User benchmark: (" + userBenchmarkName 
					+ "," + userBenchmarkResult + "," + userBenchmarkTime + ")"); 
			
			Float tmpFloat = (Float) clusterBenchmarks.get(userBenchmarkName); 
			if(tmpFloat != null) {
				float clusterResult = tmpFloat.floatValue(); 
				int clusterTime; 
				if(clusterResult != 0) {
					clusterTime = (int) (userBenchmarkTime * 
							userBenchmarkResult  / clusterResult + 0.5); 				
					clusterTimes.add(new Integer(clusterTime)); 
				}
				else {
					noMissing++; 							
				}
			}
			else { 
				noMissing++; 			
			}
		}
		
		// No user-specified benchmark advertised by cluster. 
		if(clusterTimes.size() == 0) 
			throw new Exception("No specified benchmark advertised by cluster");
		
		// Find worst execution time for the found benchmarks. 
		// Sum all times, while at it. 
		int worstTime = 0; 
		int sum = 0; 
		for(int i = 0; i < clusterTimes.size(); i++) {
			int tmp = ((Integer)clusterTimes.get(i)).intValue(); 
			sum += tmp; 
			if(worstTime < tmp)
				worstTime = tmp; 				
		}
			
		// Compensate for the non-specified benchmarks. 		
			
		// Add (worsttime + 25%) for each missing benchmark. 
		//	if(noMissing > 0)       if-clause not needed, sum += 0 is still sum
		sum += noMissing*worstTime*1.25; 								
		
		int meanTime = (int) (sum / benchmarks.length + 0.5); 		
		
		return meanTime; 
	}

    public int predict(JobDefinition_Type jobDescription, 
    		Object resource, 
			JobPreferencesType preferences, 
			Object context, 
			GlobusCredential credential) {    	
    	
    	// Method of attack: 
    	// i) First, try to predict execution time from benchmarks.
    	// ii) Should this fail, get user-specification of runtime. 
    	// iii) As a last resort, use a default job run time. 
    	
    	// i)
    	if(preferences != null) {
    		SchedulingHintsType hints = preferences.getSchedulingHints(); 
    		if(hints != null) {
    			Benchmark benchmarks[] = hints.getBenchmark();     			
    			if(benchmarks != null) {
    				if(resource instanceof GLUECERPType) {
    					try {    				
    						return benchmarkRunTime((GLUECERPType) resource, 
    								benchmarks); 
    					} catch (Exception e) {
    						logger.debug("Could not determine " 
    								+ "runtime from benchmarks"); 
    					}
    				}
    			}
    		} 
    	}
    	
    	// ii) 
    	try {    		
    		return (int) jobDescription.getJobDescription().getResources().
					getIndividualCPUTime().getExact()[0].get_value();
    	} catch (Exception e) {
    		logger.debug("Could not determine runtime from job description"); 
    	}
    	
    	// iii)     
    	int duration = Dispatcher.DEFAULT_CPU_TIME * 60; 
    	logger.debug("\"CPUTime\" not specified, assuming " + duration 
    			+ " seconds"); 
    	return duration; 
    	
    }

}
