/*
 * Created on Apr 4, 2006
 */
package se.umu.wsag.impl;

/**
 * @author tordsson
 *
 * This policy decision point is used to decide whether or not to 
 * grant a job that specifies an advance reservation identifier 
 * as an extension in the job description. 
 * 
 * Such a job is granted (and indeed run within the specified reservation), 
 * if the specified reservation exists and was created by the same
 * user that submits the job. This test is done by comparing the
 * DN in the proxy used to create the reservation with the one 
 * in the proxy used to submit the job. 
 *
 */
import java.io.File;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import javax.security.auth.Subject;
import javax.xml.namespace.QName;
import javax.xml.rpc.handler.MessageContext;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPMessage;

import org.apache.axis.message.MessageElement;
import org.apache.log4j.Logger;
import org.globus.gsi.GlobusCredential;
import org.globus.gsi.jaas.GlobusPrincipal;
import org.globus.wsrf.impl.security.authorization.exceptions.AuthorizationException;
import org.globus.wsrf.impl.security.authorization.exceptions.CloseException;
import org.globus.wsrf.impl.security.authorization.exceptions.InitializeException;
import org.globus.wsrf.impl.security.authorization.exceptions.InvalidPolicyException;
import org.globus.wsrf.security.authorization.PDP;
import org.globus.wsrf.security.authorization.PDPConfig;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public class ReservationPDP implements PDP {
	
	private String reservationProxyLocation = null; 
	
	private static Logger logger = Logger.getLogger(ReservationPDP.class); 

		 
	
	/**
	 * If job descriptions contains the element
	 * <extensions>
	 *     <ReservationType>myreservation.id.</ReservationType>
	 * </extensions>
	 * get user proxy from subject, and compare it with 
	 * the user proxy save by the AgFaSe. 	
	 * Accept job submission only if these proxys originate from 
	 * the same certificate
	 */
	public boolean isPermitted(Subject subject,
			MessageContext ctx,
			QName operation)
	throws AuthorizationException {
		
		if(1 < 2) {
			logger.warn("ReservationPDP: decision = PERMIT"); 
			return true; 
		}
		
		// We are only interested in the createManagedJob operation
		if(!"{http://www.globus.org/namespaces/2004/10/gram/job}createManagedJob".equals(operation.toString())) {
			return true; 
		}				
		
		try {
			// Extract job description (reservation id) from soap message
			org.apache.axis.MessageContext context = (org.apache.axis.MessageContext) ctx; 
			SOAPMessage message = context.getMessage();		
			SOAPBody body = message.getSOAPBody();
			
			// Get the <extensions>-tag if it exists. 								
			// Rather ugly parsing of soap element
			List allNodes = new LinkedList(); 			
			NodeList l = body.getChildNodes(); 
			for(int i = 0; i < l.getLength(); i++) {				
				MessageElement child = (MessageElement) l.item(i); 
				allNodes.add(child); 
			}
			
			String reservationId = null; 
			boolean found = false; 
			while(allNodes.size() > 0 && !found) {
				MessageElement node = (MessageElement) allNodes.remove(0); 				
				Iterator iter = node.getChildElements(); 
				while(iter.hasNext()) {
					MessageElement child = (MessageElement) iter.next(); 
					if("job".equals(child.getName())) {
						allNodes.add(child); 
					}
					if("extensions".equals(child.getName())) {
						allNodes.add(child); 
					}
					if("ReservationType".equals(child.getName())) {
						reservationId = child.getChildNodes().item(0).toString(); 
						found = true; 
					}
				}
			}
			
			if(reservationId == null) {
				logger.info("No reservation requested, decision = PERMIT"); 
				return true; 
			}	
			else {
				logger.info("Reservation with id " + reservationId + " specified:");
				
				// See if there is a proxy stored for this reservation: 							
				String filePath = this.reservationProxyLocation 
					+  "reservation." 
					+ reservationId 
					+ ".dn"; 
					//+ ".proxy";
				
				File f = new File(filePath); 
				if(!f.exists()) {
					logger.error("No reservation with id " + reservationId + " found"); 
					logger.error("decision = DENY"); 
					return false; 
				}				
				
				// Ok, user specified a valid reservation, now 
				// ensure that the user actually created this reservation. 
				
				// Get DN of caller
				Set principals = subject.getPrincipals(); 
				Iterator iter = principals.iterator(); 
				GlobusPrincipal principal = null; 
				while(iter.hasNext()) {				
					Object obj = iter.next(); 
					if(obj instanceof GlobusPrincipal) {
						principal = (GlobusPrincipal) obj; 
					}
				}
				String callerDN = principal.getName(); 
				logger.info("User with DN [" + callerDN + "] requested new job"); 
				
				// Get DN used to created referred proxy
				GlobusCredential cred = new GlobusCredential(filePath); 
				String reservationDN = cred.getIdentity(); 
					logger.info("User with DN [" + reservationDN + "] created reservation");
				
				if(callerDN.equals(reservationDN)) {
					logger.info("decision = PERMIT");
					return true; 
				}
				else {
					logger.error("decision = DENY"); 
					// TODO: Would be nice if custom error messages could be propagated 
					// back to the client, but they are not... 
					// AuthorizationException exception = new AuthorizationException(
					// 		callerDN + " does not own reservation " + reservationId); 
					// throw exception; 
					return false; 
				}
			}
		
		} catch (Exception e) {
			logger.warn("Authorization plugin failed: " + e.getMessage()); 		
			return false; 
		}
		
	}
	
	/**
	 * Read service deployment descriptor in order to set up 
	 * the location where the proxies used to created the reservations 
	 * are stored. 
	 */
	public void initialize(PDPConfig pdpConfig,
			String name,
			String id)
	throws InitializeException {
		
		// Read deployment descriptor ManagedJobFactoryService
		// and get parameter specifying reservation proxy directory. 
		Object attr = pdpConfig.getProperty(name, "proxyDirectory"); 
		
		// Use a default value just in case... 
		if(attr != null) {
			logger.info("Using proxy file location: " + attr.toString()); 
			this.reservationProxyLocation = attr.toString(); 
		}
		else {
			logger.warn("Using default proxy file location: /tmp/reservations/"); 
			this.reservationProxyLocation = "/tmp/reservations/"; 
		}
	}
	
	public void close() throws CloseException {		
	}
	
	public String[] getPolicyNames() {
		return null;
	}
	
	public Node getPolicy(Node node)
	throws InvalidPolicyException {
		return null;
	}
	
	public Node setPolicy(Node node)
	throws InvalidPolicyException {
		return null;
	}
	
	
}